import { map } from 'rxjs/operator/map';
import { RequestOptions, URLSearchParams, Headers } from '@angular/http';
import { isUndefined, stringify } from '@tdm/core/tdm';
import { findProp } from '@tdm/data';
import { UrlParamMetadata } from '../metadata/index';
import { httpDefaultConfig } from '../http-default-config';
import { getParamNames, formatPattern } from '../utils/match-pattern';
import { getHttp } from '../providers';
var HttpAdapter = (function () {
    function HttpAdapter() {
        this.supports = { cancel: true };
        this.executing = new Map();
        this.idCount = 1;
    }
    /**
     * @param {?} ctx
     * @param {?} options
     * @param {?} args
     * @return {?}
     */
    HttpAdapter.prototype.execute = function (ctx, options, args) {
        var _this = this;
        var /** @type {?} */ id = this.idCount++;
        try {
            var /** @type {?} */ http_1 = getHttp();
            if (!options)
                options = ({});
            var action = ctx.action;
            var /** @type {?} */ resource = ctx.targetMeta.model();
            if (!resource) {
                throw new Error('Http resource not set.');
            }
            var /** @type {?} */ url = findProp('endpoint', resource, action);
            if (!url) {
                // TODO: move to @tdm error with more info.
                throw new Error('Invalid endpoint, no endpoint found.');
            }
            var /** @type {?} */ withCredentials = findProp('withCredentials', httpDefaultConfig, resource, action, options);
            var /** @type {?} */ strip = findProp('trailingSlashes', httpDefaultConfig, resource, action, options);
            var /** @type {?} */ urlParams = this.getParams(ctx, ctx.targetMeta, resource, options);
            var _a = this.splitParams(url, urlParams), path = _a.path, query = _a.query;
            var /** @type {?} */ requestOptions_1 = new RequestOptions({
                url: processUrl(this.parseUrl(url, path), strip),
                method: /** @type {?} */ (action.methodInfo.source),
                search: this.paramsToSearchParams(query),
                headers: this.getHeaders(ctx, resource, options),
                body: ctx.data,
                withCredentials: withCredentials
            });
            var /** @type {?} */ response = new Promise(function (resolve, reject) {
                var /** @type {?} */ value;
                var /** @type {?} */ subscription = map.call(http_1.request(requestOptions_1.url, requestOptions_1), function (response) { return ({ data: response.json(), response: response, request: requestOptions_1 }); })
                    .subscribe(function (v) { return value = v; }, function (err) {
                    _this.executing.delete(id);
                    reject(err);
                }, function () {
                    _this.executing.delete(id);
                    resolve(value);
                });
                _this.executing.set(id, subscription);
            });
            return { id: id, response: response };
        }
        catch (err) {
            return { id: id, response: Promise.reject(err) };
        }
    };
    /**
     * @param {?} id
     * @return {?}
     */
    HttpAdapter.prototype.cancel = function (id) {
        var /** @type {?} */ sub = this.executing.get(id);
        if (sub && !sub.closed) {
            sub.unsubscribe();
        }
    };
    /**
     * @param {?} ctx
     * @param {?} resource
     * @param {?} options
     * @return {?}
     */
    HttpAdapter.prototype.getHeaders = function (ctx, resource, options) {
        var /** @type {?} */ headers = new Headers(findProp('headers', httpDefaultConfig, resource, ctx.action));
        if (options.headers) {
            Object.keys(options.headers).forEach(function (k) {
                if (isUndefined(options.headers[k])) {
                    headers.delete(k);
                }
                else {
                    headers.set(k, options.headers[k]);
                }
            });
        }
        return headers;
    };
    /**
     * @param {?} ctx
     * @param {?} meta
     * @param {?} resource
     * @param {?} options
     * @return {?}
     */
    HttpAdapter.prototype.getParams = function (ctx, meta, resource, options) {
        var /** @type {?} */ params = Object.assign({}, findProp('urlParams', httpDefaultConfig, resource, ctx.action));
        if (ctx.instance) {
            // we don't care about the keys (properties) UrlParam is on...
            // TODO: change how UrlParams are stored, instead of target->UrlParamMetadata->propName->Set<UrlParamMetadata>
            // store everything in one set/array to avoid this messy extraction.
            // an alternative is to cache the flattened array.
            var /** @type {?} */ boundParams = meta
                .getValues(UrlParamMetadata)
                .reduce(function (arr, partial) { return arr.concat(partial); }, /** @type {?} */ ([]));
            for (var /** @type {?} */ i = 0, /** @type {?} */ len = boundParams.length; i < len; i++) {
                var /** @type {?} */ bp = boundParams[i];
                if (bp.methods.length === 0 || bp.methods.some(function (mi) { return mi.method === ctx.action.method; })) {
                    params[bp.urlTemplateParamName] = ctx.instance[bp.name];
                }
            }
        }
        if (options.urlParams) {
            Object.assign(params, options.urlParams);
        }
        return params;
    };
    /**
     * @param {?} url
     * @param {?} params
     * @return {?}
     */
    HttpAdapter.prototype.parseUrl = function (url, params) {
        try {
            return formatPattern(url, params);
        }
        catch (e) {
            throw new Error("URL Parameter Error in " + stringify(this.constructor) + ": " + e.message);
        }
    };
    /**
     * From a collection of parameters return those belong to the url (path) and those that are not (query).
     * @param {?} url
     * @param {?} params
     * @return {?}
     */
    HttpAdapter.prototype.splitParams = function (url, params) {
        var /** @type {?} */ pathParamNames = getParamNames(url);
        return Object.keys(params)
            .reduce(function (splitParams, key) {
            if (!isUndefined(params[key])) {
                if (pathParamNames.indexOf(key) === -1) {
                    splitParams.query[key] = params[key];
                }
                else {
                    splitParams.path[key] = params[key];
                }
            }
            return splitParams;
        }, { path: {}, query: {} });
    };
    /**
     * @param {?} params
     * @return {?}
     */
    HttpAdapter.prototype.paramsToSearchParams = function (params) {
        return Object.keys(params)
            .reduce(function (search, key) {
            var /** @type {?} */ value = params[key];
            if (Array.isArray(value)) {
                value.forEach(function (q) { return search.append(key, q); });
            }
            else {
                search.append(key, value);
            }
            return search;
        }, new URLSearchParams());
    };
    return HttpAdapter;
}());
export { HttpAdapter };
function HttpAdapter_tsickle_Closure_declarations() {
    /** @type {?} */
    HttpAdapter.prototype.supports;
    /** @type {?} */
    HttpAdapter.prototype.executing;
    /** @type {?} */
    HttpAdapter.prototype.idCount;
}
/**
 * @param {?} url
 * @param {?} slashes
 * @return {?}
 */
function processUrl(url, slashes) {
    switch (slashes) {
        case 'strip':
            url = url.replace(/\/+$/, '') || '/';
            break;
        case 'force':
            if (url[url.length - 1] !== '/') {
                url += '/';
            }
            break;
        default:
            break;
    }
    // E.g. `http://url.com/id./format?q=x` becomes `http://url.com/id.format?q=x`
    return url.replace(/\/\.(?=\w+($|\?))/, '.')
        .replace(/\/\\\./, '/.'); // replace escaped `/\.` with `/.`
}
//# sourceMappingURL=http-adapter.js.map