// Taken from npm module 'path-to-regexp'
// couldn't import to to rollup issues with commonjs plugin
/**
 * The main path matching regexp utility.
 *
 */
var PATH_REGEXP = new RegExp([
    // Match escaped characters that would otherwise appear in future matches.
    // This allows the user to escape special characters that won't transform.
    '(\\\\.)',
    // Match Express-style parameters and un-named parameters with a prefix
    // and optional suffixes. Matches appear as:
    //
    // "/:test(\\d+)?" => ["/", "test", "\d+", undefined, "?", undefined]
    // "/route(\\d+)"  => [undefined, undefined, undefined, "\d+", undefined, undefined]
    // "/*"            => ["/", undefined, undefined, undefined, undefined, "*"]
    '([\\/.])?(?:(?:\\:(\\w+)(?:\\(((?:\\\\.|[^\\\\()])+)\\))?|\\(((?:\\\\.|[^\\\\()])+)\\))([+*?])?|(\\*))'
].join('|'), 'g');
/**
 * Parse a string for the raw tokens.
 *
 * @param {?} str
 * @param {?} options
 * @return {?}
 */
export function parse(str, options) {
    var /** @type {?} */ tokens = [];
    var /** @type {?} */ key = 0;
    var /** @type {?} */ index = 0;
    var /** @type {?} */ path = '';
    var /** @type {?} */ defaultDelimiter = options && options.delimiter || '/';
    var /** @type {?} */ res;
    while ((res = PATH_REGEXP.exec(str)) != null) {
        var /** @type {?} */ m = res[0];
        var /** @type {?} */ escaped = res[1];
        var /** @type {?} */ offset = res.index;
        path += str.slice(index, offset);
        index = offset + m.length;
        // Ignore already escaped sequences.
        if (escaped) {
            path += escaped[1];
            continue;
        }
        var /** @type {?} */ next = str[index];
        var /** @type {?} */ prefix = res[2];
        var /** @type {?} */ name = res[3];
        var /** @type {?} */ capture = res[4];
        var /** @type {?} */ group = res[5];
        var /** @type {?} */ modifier = res[6];
        var /** @type {?} */ asterisk = res[7];
        // Push the current path onto the tokens.
        if (path) {
            tokens.push(path);
            path = '';
        }
        var /** @type {?} */ partial = prefix != null && next != null && next !== prefix;
        var /** @type {?} */ repeat = modifier === '+' || modifier === '*';
        var /** @type {?} */ optional = modifier === '?' || modifier === '*';
        var /** @type {?} */ delimiter = res[2] || defaultDelimiter;
        var /** @type {?} */ pattern = capture || group;
        tokens.push({
            name: name || key++,
            prefix: prefix || '',
            delimiter: delimiter,
            optional: optional,
            repeat: repeat,
            partial: partial,
            asterisk: !!asterisk,
            pattern: pattern ? escapeGroup(pattern) : (asterisk ? '.*' : '[^' + escapeString(delimiter) + ']+?')
        });
    }
    // Match any characters still remaining.
    if (index < str.length) {
        path += str.substr(index);
    }
    // If the path exists, push it onto the end.
    if (path) {
        tokens.push(path);
    }
    return tokens;
}
/**
 * Compile a string to a template function for the path.
 *
 * @param {?} str
 * @param {?=} options
 * @return {?}
 */
export function compile(str, options) {
    return tokensToFunction(parse(str, options));
}
/**
 * Prettier encoding of URI path segments.
 *
 * @param {?} str
 * @return {?}
 */
function encodeURIComponentPretty(str) {
    return encodeURI(str).replace(/[\/?#]/g, function (c) {
        return '%' + c.charCodeAt(0).toString(16).toUpperCase();
    });
}
/**
 * Encode the asterisk parameter. Similar to `pretty`, but allows slashes.
 *
 * @param {?} str
 * @return {?}
 */
function encodeAsterisk(str) {
    return encodeURI(str).replace(/[?#]/g, function (c) {
        return '%' + c.charCodeAt(0).toString(16).toUpperCase();
    });
}
/**
 * Expose a method for transforming tokens into the path function.
 * @param {?} tokens
 * @return {?}
 */
export function tokensToFunction(tokens) {
    // Compile all the tokens into regexps.
    var /** @type {?} */ matches = new Array(tokens.length);
    // Compile all the patterns before compilation.
    for (var /** @type {?} */ i = 0; i < tokens.length; i++) {
        if (typeof tokens[i] === 'object') {
            matches[i] = new RegExp('^(?:' + tokens[i].pattern + ')$');
        }
    }
    return function (obj, opts) {
        var /** @type {?} */ path = '';
        var /** @type {?} */ data = obj || {};
        var /** @type {?} */ options = opts || {};
        var /** @type {?} */ encode = options.pretty ? encodeURIComponentPretty : encodeURIComponent;
        for (var /** @type {?} */ i = 0; i < tokens.length; i++) {
            var /** @type {?} */ token = tokens[i];
            if (typeof token === 'string') {
                path += token;
                continue;
            }
            var /** @type {?} */ value = data[token.name];
            var /** @type {?} */ segment;
            if (value == null) {
                if (token.optional) {
                    // Prepend partial segment prefixes.
                    if (token.partial) {
                        path += token.prefix;
                    }
                    continue;
                }
                else {
                    throw new TypeError('Expected "' + token.name + '" to be defined');
                }
            }
            if (Array.isArray(value)) {
                if (!token.repeat) {
                    throw new TypeError('Expected "' + token.name + '" to not repeat, but received `' + JSON.stringify(value) + '`');
                }
                if (value.length === 0) {
                    if (token.optional) {
                        continue;
                    }
                    else {
                        throw new TypeError('Expected "' + token.name + '" to not be empty');
                    }
                }
                for (var /** @type {?} */ j = 0; j < value.length; j++) {
                    segment = encode(value[j]);
                    if (!matches[i].test(segment)) {
                        throw new TypeError('Expected all "' + token.name + '" to match "' + token.pattern + '", but received `' + JSON.stringify(segment) + '`');
                    }
                    path += (j === 0 ? token.prefix : token.delimiter) + segment;
                }
                continue;
            }
            segment = token.asterisk ? encodeAsterisk(value) : encode(value);
            if (!matches[i].test(segment)) {
                throw new TypeError('Expected "' + token.name + '" to match "' + token.pattern + '", but received "' + segment + '"');
            }
            path += token.prefix + segment;
        }
        return path;
    };
}
/**
 * Escape a regular expression string.
 *
 * @param {?} str
 * @return {?}
 */
function escapeString(str) {
    return str.replace(/([.+*?=^!:${}()[\]|\/\\])/g, '\\$1');
}
/**
 * Escape the capturing group by escaping special characters and meaning.
 *
 * @param {?} group
 * @return {?}
 */
function escapeGroup(group) {
    return group.replace(/([=!:$\/()])/g, '\\$1');
}
/**
 * Attach the keys as a property of the regexp.
 *
 * @param {?} re
 * @param {?} keys
 * @return {?}
 */
function attachKeys(re, keys) {
    re.keys = keys;
    return re;
}
/**
 * Get the flags for a regexp from the options.
 *
 * @param {?} options
 * @return {?}
 */
function flags(options) {
    return options.sensitive ? '' : 'i';
}
/**
 * Pull out keys from a regexp.
 *
 * @param {?} path
 * @param {?} keys
 * @return {?}
 */
function regexpToRegexp(path, keys) {
    // Use a negative lookahead to match only capturing groups.
    var /** @type {?} */ groups = path.source.match(/\((?!\?)/g);
    if (groups) {
        for (var /** @type {?} */ i = 0; i < groups.length; i++) {
            keys.push({
                name: i,
                prefix: null,
                delimiter: null,
                optional: false,
                repeat: false,
                partial: false,
                asterisk: false,
                pattern: null
            });
        }
    }
    return attachKeys(path, keys);
}
/**
 * Transform an array into a regexp.
 *
 * @param {?} path
 * @param {?} keys
 * @param {?} options
 * @return {?}
 */
function arrayToRegexp(path, keys, options) {
    var /** @type {?} */ parts = [];
    for (var /** @type {?} */ i = 0; i < path.length; i++) {
        parts.push(pathToRegexp(path[i], keys, options).source);
    }
    var /** @type {?} */ regexp = new RegExp('(?:' + parts.join('|') + ')', flags(options));
    return attachKeys(regexp, keys);
}
/**
 * Create a path regexp from string input.
 *
 * @param {?} path
 * @param {?} keys
 * @param {?} options
 * @return {?}
 */
function stringToRegexp(path, keys, options) {
    return tokensToRegExp(parse(path, options), keys, options);
}
/**
 * Expose a function for taking tokens and returning a RegExp.
 *
 * @param {?} tokens
 * @param {?} keys
 * @param {?} options
 * @return {?}
 */
export function tokensToRegExp(tokens, keys, options) {
    if (!Array.isArray(keys)) {
        options = (keys || options);
        keys = [];
    }
    options = options || {};
    var /** @type {?} */ strict = options.strict;
    var /** @type {?} */ end = options.end !== false;
    var /** @type {?} */ route = '';
    // Iterate over the tokens and create our regexp string.
    for (var /** @type {?} */ i = 0; i < tokens.length; i++) {
        var /** @type {?} */ token = tokens[i];
        if (typeof token === 'string') {
            route += escapeString(token);
        }
        else {
            var /** @type {?} */ prefix = escapeString(token.prefix);
            var /** @type {?} */ capture = '(?:' + token.pattern + ')';
            keys.push(token);
            if (token.repeat) {
                capture += '(?:' + prefix + capture + ')*';
            }
            if (token.optional) {
                if (!token.partial) {
                    capture = '(?:' + prefix + '(' + capture + '))?';
                }
                else {
                    capture = prefix + '(' + capture + ')?';
                }
            }
            else {
                capture = prefix + '(' + capture + ')';
            }
            route += capture;
        }
    }
    var /** @type {?} */ delimiter = escapeString(options.delimiter || '/');
    var /** @type {?} */ endsWithDelimiter = route.slice(-delimiter.length) === delimiter;
    // In non-strict mode we allow a slash at the end of match. If the path to
    // match already ends with a slash, we remove it for consistency. The slash
    // is valid at the end of a path match, not in the middle. This is important
    // in non-ending mode, where "/test/" shouldn't match "/test//route".
    if (!strict) {
        route = (endsWithDelimiter ? route.slice(0, -delimiter.length) : route) + '(?:' + delimiter + '(?=$))?';
    }
    if (end) {
        route += '$';
    }
    else {
        // In non-ending mode, we need the capturing groups to match as much as
        // possible by using a positive lookahead to the end or next path segment.
        route += strict && endsWithDelimiter ? '' : '(?=' + delimiter + '|$)';
    }
    return attachKeys(new RegExp('^' + route, flags(options)), keys);
}
/**
 * Normalize the given path string, returning a regular expression.
 *
 * An empty array can be passed in for the keys, which will hold the
 * placeholder key descriptions. For example, using `/user/:id`, `keys` will
 * contain `[{ name: 'id', delimiter: '/', optional: false, repeat: false }]`.
 *
 * @param {?} path
 * @param {?=} keys
 * @param {?=} options
 * @return {?}
 */
export function pathToRegexp(path, keys, options) {
    if (!Array.isArray(keys)) {
        options = (keys || options);
        keys = [];
    }
    options = options || {};
    if (path instanceof RegExp) {
        return regexpToRegexp(path, /** @type {!Array} */ (keys));
    }
    if (Array.isArray(path)) {
        return arrayToRegexp(/** @type {!Array} */ (path), /** @type {!Array} */ (keys), options);
    }
    return stringToRegexp(/** @type {string} */ (path), /** @type {!Array} */ (keys), options);
}
//# sourceMappingURL=path-to-regexp.js.map